/*
 * jabberoo-browse.cpp
 * Jabber Client Library Browse Support
 *
 * Original Code Copyright (C) 1999-2001 Dave Smith (dave@jabber.org)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contributor(s): Julian Missig (IBM)
 *
 */

#include "jabberoo.hh"

#include <sigc++/object_slot.h>

using namespace judo;
using namespace jabberoo;

DiscoDB::Item::Identity::Identity(const Element& e)
{
    _type = e.getAttrib("type");
    _category = e.getAttrib("category");
}

DiscoDB::Item::Identity::~Identity()
{ }

DiscoDB::Item::~Item()
{
    while(!_identities.empty())
    {
        DiscoDB::Item::Identity* ident = _identities.front();
        delete ident;
        _identities.pop_front();
    }
}

DiscoDB::DiscoDB(Session& sess) : 
    _session(sess) 
{
}

DiscoDB::~DiscoDB()
{
    clear();
}

DiscoDB::Item& DiscoDB::operator[](const std::string& jid)
{
    DiscoDB::iterator it = _items.find(jid);
    if (it == _items.end())
    {
        throw XCP_NotCached();
    }

    return *(it->second);
}

void DiscoDB::cache(const std::string& jid, DiscoCallbackFunc f)
{
    // Hook up the callback
    _callbacks.insert(CallbackMap::value_type(jid, f));
    
    // Build our node
    judo::Element iq("iq");
    iq.putAttrib("type", "get");
    std::string id = _session.getNextID();
    iq.putAttrib("id", id);
    iq.putAttrib("to", jid);
    judo::Element* query = iq.addElement("query");
    query->putAttrib("xmlns", "http://jabber.org/protocol/disco#items");

    _session.registerIQ(id, SigC::slot(*this, &DiscoDB::discoItemsCB));

    // Send it out
    _session << iq.toString().c_str();

    // Dupe checking and what not is handled by the xpath that picks up results
}

void DiscoDB::cache(const std::string& jid, const std::string& node,
    DiscoCallbackFunc f)
{
    // Hook up the callback
    _callbacks.insert(CallbackMap::value_type(jid + "::" + node, f));
    
    // Build our node
    judo::Element iq("iq");
    iq.putAttrib("type", "get");
    std::string id = _session.getNextID();
    iq.putAttrib("id", id);
    iq.putAttrib("to", jid);
    judo::Element* query = iq.addElement("query");
    query->putAttrib("xmlns", "http://jabber.org/protocol/disco#items");
    query->putAttrib("node", node);

    _session.registerIQ(id, SigC::slot(*this, &DiscoDB::discoInfoCB));

    // Send it out
    _session << iq.toString().c_str();
}

void DiscoDB::clear()
{
    for (DiscoDB::iterator it = _items.begin(); it != _items.end(); ++it)
    {
        delete it->second;
    }

    _items.clear();
}

void DiscoDB::discoInfoCB(const Element& e)
{
    std::string jid = e.getAttrib("from");

    // Make sure we already got the items stuff in there
    pair<ItemMap::iterator, ItemMap::iterator> items = _items.equal_range(jid);
    if (items.first == items.second)
    {
        runCallbacks(jid, NULL);
        return;
    }
    DiscoDB::Item* item = NULL;
    
    const Element* query = e.findElement("query");

    // If we are working on a specific node we need to pull it out
    std::string node = query ? query->getAttrib("node") : "";

    // Find the one with the same node value
    for (ItemMap::iterator i = items.first; i != items.second; ++i)
    {
        if (i->second->getNode() == node)
        {
            item = i->second;
            break;
        }
    }
    
    // Now we can add the identities and features
    for (Element::const_iterator it = query->begin(); it != query->end(); ++it)
    {
        if ((*it)->getType() == Node::ntElement)
        {
            Element* elem = static_cast<Element*>((*it));
            std::string name = elem->getName();
            if (name == "feature")
            {
                std::string feature = elem->getAttrib("var");
                if (!feature.empty())
                    item->addFeature(feature);
            }
            else if (name == "identity")
            {
                item->addIdentity(*elem);
            }
        }
    }

    runCallbacks(jid, item);
}

void DiscoDB::discoItemsCB(const Element& e)
{
    std::string type = e.getAttrib("type");
    std::string jid = e.getAttrib("from");

    // Catch errors and see if we can browse instead
    if (type == "error")
    {
        judo::Element iq("iq");
        iq.putAttrib("type", "get");
        std::string id = _session.getNextID();
        iq.putAttrib("id", id);
        iq.putAttrib("to", jid);
        judo::Element* item_query = iq.addElement("item");
        item_query->putAttrib("xmlns", "jabber:iq:browse");

        _session.registerIQ(id, SigC::slot(*this, &DiscoDB::browseCB));

        _session << iq;

        return;
    }

    DiscoDB::Item *item = NULL;

    // Make our new item to insert
    item = new DiscoDB::Item(jid);

    // Clean up old entries since we are now the newest version
    pair<ItemMap::iterator, ItemMap::iterator> items = _items.equal_range(jid);
    _items.erase(items.first, items.second);

    // They don't have us at all, just pop them in
    _items.insert(ItemMap::value_type(jid, item));
    
    // Add any children that are there
    const Element* query = e.findElement("query");
    std::string node = query ? query->getAttrib("node") : "";
    if (!node.empty())
        item->setNode(node);

    if (query)
    {
        for (Element::const_iterator it = query->begin(); 
             it != query->end(); ++it)
        {
            if ( (*it)->getType() != Node::ntElement )
            {
                continue;
            }

            // Build the child and stick it in
            Element* elem = static_cast<Element*>(*it);
            if (elem->getName() == "item")
            {
                std::string cjid = elem->getAttrib("jid");
                std::string cname = elem->getAttrib("name");
                std::string cnode = elem->getAttrib("node");
                DiscoDB::Item* child = NULL;
                DiscoDB::iterator nit = _items.find(cjid);
                if (nit == _items.end())
                {
                    child = new DiscoDB::Item(cjid);
                    if (!cname.empty())
                        child->setName(cname);
                    if (!cnode.empty())
                        child->setNode(cnode);
                    _items.insert(ItemMap::value_type(cjid, child));
                    item->appendChild(child);
                }
                else
                {
                    // XXX Do something more here?  Possibly update name?
                }
            }
        }
    }
            
    // Now we need the info element
    judo::Element iq("iq");
    iq.putAttrib("type", "get");
    std::string id = _session.getNextID();
    iq.putAttrib("id", id);
    iq.putAttrib("to", jid);
    judo::Element* info_query = iq.addElement("query");
    info_query->putAttrib("xmlns", "http://jabber.org/protocol/disco#info");
    if (!node.empty())
        info_query->putAttrib("node", node);

    _session.registerIQ(id, SigC::slot(*this, &DiscoDB::discoInfoCB));

    // Send it out
    _session << iq.toString().c_str();
}

void DiscoDB::browseCB(const Element& e)
{
    Element* child = NULL;
    std::string jid;

    // Find the first child element and ensure it has a jid
    for (Element::const_iterator it = e.begin(); it != e.end(); ++it)
    {
        if ( (*it)->getType() == Node::ntElement )
        {
           child = static_cast<Element*>(*it);
           jid = child->getAttrib("jid");
           if (!jid.empty())
               break;
        }
    }

    if (child == NULL)
    {
        // XXX MAJOR ERROR HERE
        return;
    }

    // We need to make sure we don't already have a copy of this, if we do
    // clean it up and let the new one take precedence
    DiscoDB::iterator it = _items.find(jid);
    if (it != _items.end())
    {
        _items.erase(it);
    }

    // Cache it up
    DiscoDB::Item* item = new DiscoDB::Item(jid);
    // Now we add all the children and sneak in there identities
    Element* identity;
    for (Element::iterator i = child->begin(); i != child->end(); ++i)
    {
        if ( (*i)->getType() != Node::ntElement )
        {
            continue;
        }

        identity = static_cast<Element*>(*i);
        std::string cjid = identity->getAttrib("jid");
        // This would be fubar anyway
        if (cjid.empty())
            continue;

        DiscoDB::Item* citem = new DiscoDB::Item(cjid);
        citem->setName(identity->getAttrib("name"));
        item->appendChild(citem);
        std::string cat = identity->getAttrib("category");
        // See if it's old skool
        if (cat.empty())
            cat = citem->getName();
        std::string type = identity->getAttrib("type");
        citem->addIdentity(cat, type);
        // See if we have namespaces to fake as features
        for (Element::iterator nsi = identity->begin();
             nsi != identity->end(); ++nsi)
        {
            if ( ((*i)->getType() == Node::ntElement) &&
                 ((*i)->getName() == "ns") )
            {
                judo::Element* ns = static_cast<Element*>(*i);
                citem->addFeature(ns->getCDATA());
            }
        }
    }
    _items.insert(ItemMap::value_type(jid, item));

    runCallbacks(jid, item);
}

void DiscoDB::runCallbacks(const std::string& jid, const DiscoDB::Item* item)
{
    // See if we have a callback for this lookup
    std::pair<CallbackMap::iterator, CallbackMap::iterator> cis =
        _callbacks.equal_range(jid);

    if (cis.first == cis.second)
        return;

    // XXX this could be a for_each
    for(CallbackMap::iterator i = cis.first; i != cis.second; ++i)
    {
        i->second(item);
        _callbacks.erase(i);
    }
}
